#include <iomanip>
#include <iostream>
#include <fstream>
#include <string>
#include <sstream>
#include <cmath>
#include <cstdlib>
//#include <cuba.h>
#include <gsl/gsl_errno.h>

#include "CoefficientFunctions.hh"
#include "CoefficientFunctionsOffShell.hh"
#include "Pqg.hh"
#include "math/borel.hh"
#include "integration.hh"
#include "gammaNLO.hh"
#include "h_coefficients.hh"
//#include <complex_bessel/sp_airy.h>


using namespace std;



namespace HELL {

  const double CA = 3.;


  double ArgBC(dcomplex z) {
    double res = arg(z);
    //if(res<0) res += 2*M_PI;
    //if(res>0) res -= 2*M_PI;
    return res;
  }
  dcomplex LogBC(dcomplex z) {
    return log(abs(z)) + I*ArgBC(z);
  }
  dcomplex evolU(double xi, dcomplex g, dcomplex r) {
    if(xi==0) return 0.;
    dcomplex res = exp( g/r * LogBC(r*log(xi)+1.) );
    //if(imag(r)<0 && real(r)*log(xi)+1.<0) res *= exp( g/r * I*2.*M_PI);
    //if(imag(r)<0 && real(r)*log(xi)+1.<0) res = exp( g/r * (log(r*log(xi)+1.)+I*2.*M_PI) );
    return res;
  }


  
  dcomplex integrand(double xi, dcomplex gamma, dcomplex r, process p, double x1) {
    dcomplex U = evolU(xi,gamma,r);
    dcomplex res = U;
    //
    if(p == gamma_qg) res *= - Dxi_Hqg(xi);
    //if(p == CL) res *= - Dxi_CLoff(xi, x1);
    //if(p == C2) res *= - Dxi_C2off(xi, x1) - (xi<=1 ? 1./xi : 0);
    if(p == CL) res *= - Dxi_CLoff(xi);
    if(p == C2) res *= - Dxi_C2off(xi) - (xi<=1 ? 1./xi : 0);
    //
    //cout << setw(15) << xi << setw(15) << x1 << setw(35) << res << endl;
    return res;
  }
  struct s_pr2c2d { process p; dcomplex c1, c2; double d3, d4; };
  /*
  int integrandCUBA(const int *ndim, const double x[], const int *ncomp, double res[], void *pars) {
    s_pr2c2d p = *(s_pr2c2d*)pars;
    dcomplex g  = p.c1;
    dcomplex r  = p.c2;
    double a0 = p.d3;  // as(Q^2)
    double b0 = p.d4;  // beta0
    //double xi0 = real(r); // may be negative - problematic
    double xi0 = a0*b0;
    //double xi = exp(-1./a0/b0) + (1.-x[0])/x[0];  double jac = 1./x[0]/x[0];
    //double amin = 0.01*a0, amax = 2000*a0;  double a = amin + (amax-amin)*x[0];  double xi = exp((1/a-1/a0)/b0);  double jac = -xi/(-b0*a*a) * (amax-amin);
    //double Xmin = 0.1, Xmax = 80;  double X = Xmin + (Xmax-Xmin)*x[0];  double xi = exp((1-X)/X/xi0);  double jac = xi/X/X/xi0*(Xmax-Xmin);
    double Xmin = 0.1, Xmax = 80;  double X1 = Xmin + (1-Xmin)*x[0];  double xi1 = exp((1-X1)/X1/xi0);  double jac1 = xi1/X1/X1/xi0*(1-Xmin); double X2 = 1 + (Xmax-1)*(1-x[0]);  double xi2 = exp((1-X2)/X2/xi0);  double jac2 = xi2/X2/X2/xi0*(Xmax-1);
    //double Xmin = 0.1;  double X1 = Xmin + (1-Xmin)*x[0];  double xi1 = exp((1-X1)/X1/xi0);  double jac1 = xi1/X1/X1/xi0*(1-Xmin);  double X2 = x[0];  double xi2 = exp((1-1/X2)*X2/xi0);  double jac2 = xi2/xi0;
    //
    //dcomplex result = jac * integrand(xi, g, r, p.p, x[1]);
    dcomplex result = jac1 * integrand(xi1, g, r, p.p, x[1]) + jac2 * integrand(xi2, g, r, p.p, x[1]);
    //
    //cout << setw(15) << xi << setw(15) << x[1] << setw(35) << res << endl;
    res[0] = real(result);
    res[1] = imag(result);
    return 0;
  }
  */
  dcomplex integrandGSL(double x, void *pars) {
    s_pr2c2d p = *(s_pr2c2d*)pars;
    dcomplex g  = p.c1;
    dcomplex r  = p.c2;
    double a0 = p.d3;  // as(Q^2)
    double b0 = p.d4;  // beta0
    //double xi0 = real(r); // may be negative - problematic
    double xi0 = a0*b0;
    //double xi = exp(-1./a0/b0) + (1.-x)/x;  double jac = 1./x/x;
    //double amin = 0.01*a0, amax = 2000*a0;  double a = amin + (amax-amin)*x;  double xi = exp((1/a-1/a0)/b0);  double jac = -xi/(-b0*a*a) * (amax-amin);
    //double Xmin = 0.1, Xmax = 80;  double X = Xmin + (Xmax-Xmin)*x;  double xi = exp((1-X)/X/xi0);  double jac = xi/X/X/xi0*(Xmax-Xmin);
    double Xmin = 0.05, Xmax = 70;  double X1 = Xmin + (1-Xmin)*x;  double xi1 = exp((1-X1)/X1/xi0);  double jac1 = xi1/X1/X1/xi0*(1-Xmin); double X2 = 1 + (Xmax-1)*(1-x);  double xi2 = exp((1-X2)/X2/xi0);  double jac2 = xi2/X2/X2/xi0*(Xmax-1);
    //double Xmin = 0.1;  double X1 = Xmin + (1-Xmin)*x;  double xi1 = exp((1-X1)/X1/xi0);  double jac1 = xi1/X1/X1/xi0*(1-Xmin);  double X2 = x;  double xi2 = exp((1-1/X2)*X2/xi0);  double jac2 = xi2/xi0;
    //cout << x << "  " << xi << endl;
    //if(1+real(r)*log(xi2)<0) cout << "warning!" << endl;
    //
    //dcomplex result = jac * integrand(xi, g, r, p.p, 0);
    dcomplex result = jac1 * integrand(xi1, g, r, p.p, 0) + jac2 * integrand(xi2, g, r, p.p, 0);
    //
    //cout << setw(15) << xi << setw(15) << x[1] << setw(35) << res << endl;
    return result;
  }
  double integrandGSLre(double x, void *p) {
    return real(integrandGSL(x,p));
  }
  double integrandGSLim(double x, void *p) {
    return imag(integrandGSL(x,p));
  }

  //*
#include "hypergU.cc"
  dcomplex Bateman_logderiv(dcomplex nu, dcomplex x) {
    dcomplex num = hypergU(1.-nu*0.5,1.,2.*x);
    dcomplex den = hypergU(-nu*0.5,0.,2.*x);
    // Sometimes den is too large, and its norm can be inf.
    // Since in num/den the norm(den) is needed, in those cases num/den returns a NaN.
    // This renormalization solves the problem.
    if(real(den)!=0) {
      num /= real(den);
      den /= real(den);
    }
    //cout << num << "  " << den << endl;
    //dcomplex res = -1. + nu * num/den;
    dcomplex res = -1. + nu * exp(log(num)-log(den));
    //
    return res;
  }
  dcomplex gammas_run(dcomplex N, double as, double nf) {
    double b0 = beta0(nf);
    double c0 = CA/M_PI * 4*log(2.);
    double c = c0 - 4.*CA/M_PI;
    //c += CA/M_PI*(4-7./2*ZETA3);
    dcomplex A = (1./as-c/N)/b0;
    dcomplex m2x = 2.*CA/M_PI/b0/N;
    //m2x *= 14.*ZETA3/16.;
    dcomplex res = 1./2. - 1./A + 1./2.*Bateman_logderiv(m2x,A/2.);
    return res + b0*as;
  }
  //*/
  /*
  inline dcomplex var_z(dcomplex N, double as, double c, double kappa, double nf) {
    return pow(2*beta0(nf)*N*as/kappa, 1./3.) * (1. - c/N) /as/beta0(nf);
  }
  dcomplex gamma_Airy(dcomplex N, double as, double c, double kappa, double nf) {
    dcomplex z = var_z(N,as,c,kappa,nf);
    sp_airy  Airy(0, 0);
    sp_airy DAiry(0, 1);
    dcomplex logDer = DAiry(z) / Airy(z);
    return 0.5 + pow(2*beta0(nf)*N*as/kappa, 1./3.) * logDer;
  }
  dcomplex gamma_Airy_asympt(dcomplex N, double as, double c, double kappa, double nf) {
    return 0.5 - sqrt( 2.*(N-c)/kappa ) -as*beta0(nf)/4.;
  }
  dcomplex gamma_Airy_minus_asympt(dcomplex N, double as, double c, double kappa, double nf) {
    return gamma_Airy(N, as, c, kappa, nf) - gamma_Airy_asympt(N, as, c, kappa, nf);
  }
  dcomplex gammas(dcomplex N, double as, double nf) {
    dcomplex K = M_PI/CA*N/as - 4.*(log(2.)-1.);
    return 0.5*( 1.-sqrt(1.-4./K) );
  }
  dcomplex gammas_Airy_matched(dcomplex N, double as, double nf) {
    double b0 = beta0(nf);
    double c0 = CA/M_PI * 4*log(2.);
    double c     = as*c0;
    double kappa = CA/M_PI*as * 32.;
    dcomplex res = gamma_Airy_minus_asympt(N, as, c, kappa, nf) + gammas(N,as,nf);
    return res + b0*as/4.;
  }
  */


  dcomplex newDeltaFRes(dcomplex Delta_gammaLO_BatNLO, dcomplex Delta_gammaLOdot_BatNLO, dcomplex N, double as, double nf, process proc, int ordersub=0) {
    hCoeff hC(proc);
    double h0 = hC.hk(0);
    double h1 = hC.hk(1);
    //
    double b0 = beta0(nf);
    dcomplex fixedorderpole = gamma0sing(N,nf);
    //fixedorderpole = gamma0(N,nf) - gamma0(N+1.,nf);
    //fixedorderpole = gamma0(N,nf) - 2.*gamma0(N+1.,nf) + gamma0(N+2.,nf);
    //fixedorderpole = gamma0(N,nf) - 3.*gamma0(N+1.,nf) + 3.*gamma0(N+2.,nf) - gamma0(N+3.,nf);
    //
    //*
    // Marco version
    dcomplex gammaLO_BatNLO    = Delta_gammaLO_BatNLO    + as*fixedorderpole;//  + as*as*gamma1sing(N,nf);
    dcomplex gammaLOdot_BatNLO = Delta_gammaLOdot_BatNLO +    fixedorderpole;
    //
    //gammaLO_BatNLO    = gammas_run(N,as,nf); //hack!!!
    //gammaLO_BatNLO    = gammas_Airy_matched(N,as,nf); //hack!!!
    //gammaLO_BatNLO   *= 2./(N+2.); //hack!!!
    //gammaLO_BatNLO /= N+1.; // hack! large-N damping
    //gammaLOdot_BatNLO /= N+1.; // hack! large-N damping
    //gammaLOdot_BatNLO = gammaLO_BatNLO/as; //hack!!!
    //
    gammaLOdot_BatNLO *= -b0*as*as;  // convert deriv wrt as to deriv wrt t
    dcomplex r = -gammaLOdot_BatNLO / gammaLO_BatNLO;
    //
    s_pr2c2d pars = { proc, gammaLO_BatNLO, r, as, b0 };
    dcomplex res;
    //
    // const int ncomp = 2;
    // double cres[ncomp], error[ncomp], prob[ncomp];
    // double epsrel=1e-3, epsabs=1.e-10;
    // //if(proc == C2) epsrel /= 3;
    // int last = 4;
    // int verbose = 0;
    // int nregions, neval, fail;
    // int mineval = 0, maxeval = 500000;
    // //nregions=1; // to remove the warning
    // int dim = 2;
    // //
    // Cuhre(dim, ncomp, integrandCUBA, &pars, 1, epsrel, epsabs, verbose | last, mineval, maxeval, 9, NULL, NULL, &nregions, &neval, &fail, cres, error, prob);
    // //
    // if(fail == -1)  cout << "  ERROR: dimension out of range." << endl;
    // else if(fail>0) cout << "  ERROR: precision not reached." << endl;
    // //if(verbosity) cout << "Integral result: " << setw(12) << cres[0] << " +/- " << setw(12) << error[0] << "  (relative error " << setw(12) << error[0]/cres[0] <<")" << endl;
    // //
    // res = cres[0] + I*cres[1];
    // //cout << res << endl;
    // //if(res != 1.) exit(0);
    //
    gsl_error_handler_t *old_handler = gsl_set_error_handler (NULL);
    gsl_set_error_handler_off();
    double prec = 1e-4;
    res = integration_gauss(integrandGSLre, 0, 1, prec, NULL, &pars)
      + I*integration_gauss(integrandGSLim, 0, 1, prec, NULL, &pars);
    gsl_set_error_handler (old_handler);
    /*/
    // Richard version
    dcomplex FOgammaLO  = as*gamma0pl(N+1.,0);
    dcomplex gammaLO_BatNLO    = Delta_gammaLO_BatNLO    + FOgammaLO;
    dcomplex gammaLOdot_BatNLO = Delta_gammaLOdot_BatNLO + FOgammaLO/as;
    gammaLOdot_BatNLO *= -b0*as*as;  // convert deriv wrt as to deriv wrt t
    FOgammaLO -= as*gamma0sing(N,0);
    dcomplex FOgammaLOdot  = -b0*as*FOgammaLO;
    //gammaLOdot_BatNLO = 0;
    //FOgammaLOdot = 0;
    //
    dcomplex r   = -gammaLOdot_BatNLO / gammaLO_BatNLO;
    dcomplex rLO = -FOgammaLOdot / FOgammaLO;
    //
    s_pr2c2d pars   = { proc,   gammaLO_BatNLO, r,   as, b0 };
    s_pr2c2d parsLO = { proc, FOgammaLO,        rLO, as, b0 };
    dcomplex res;
    gsl_error_handler_t *old_handler = gsl_set_error_handler (NULL);
    gsl_set_error_handler_off();
    double prec = 1e-4;
    res  = integration_gauss(integrandGSLre, 0, 1, prec, NULL, &pars)
      +  I*integration_gauss(integrandGSLim, 0, 1, prec, NULL, &pars);
    res -= integration_gauss(integrandGSLre, 0, 1, prec, NULL, &parsLO)
      +  I*integration_gauss(integrandGSLim, 0, 1, prec, NULL, &parsLO);
    res += h0;
    gsl_set_error_handler (old_handler);
    //*/
    //
    //if(proc == gamma_qg) res += evolU(1.,gammaLO_BatNLO,r) * Dxi_Hqg(1.);
    //
    if(ordersub>0) res -= h0;
    if(ordersub>1) res -= h1*as*fixedorderpole;
    return res;
  }


  dcomplex newDeltaGammaResNLOqg(dcomplex Delta_gammaLO_BatNLO, dcomplex Delta_gammaLOdot_BatNLO, dcomplex N, double as, double nf, dcomplex Delta_gammaNLO) {
    dcomplex res = newDeltaFRes(Delta_gammaLO_BatNLO, Delta_gammaLOdot_BatNLO, N, as, nf, gamma_qg, 2);
    return res * as*nf/3./M_PI;
  }
  dcomplex DeltaCLRes(dcomplex Delta_gammaLO_BatNLO, dcomplex Delta_gammaLOdot_BatNLO, dcomplex N, double as, double nf, dcomplex Delta_gammaNLO) {
    dcomplex res = newDeltaFRes(Delta_gammaLO_BatNLO, Delta_gammaLOdot_BatNLO, N, as, nf, CL, 2);
    return res * as*nf/3./M_PI;
    //return res * as/3./M_PI;
  }
  dcomplex DeltaC2Res(dcomplex Delta_gammaLO_BatNLO, dcomplex Delta_gammaLOdot_BatNLO, dcomplex N, double as, double nf, dcomplex Delta_gammaNLO) {
    dcomplex res = newDeltaFRes(Delta_gammaLO_BatNLO, Delta_gammaLOdot_BatNLO, N, as, nf, C2, 2);
    res *= as*nf/3./M_PI;
    //res *= as/3./M_PI;
    res -= DeltaUqg(Delta_gammaLO_BatNLO, Delta_gammaLOdot_BatNLO, N, as, nf, Delta_gammaNLO);
    return res;
  }






};

